__all__ = (
    "safety_masks",
    "standard_bf",
    "gen_long_gaps_survey",
    "gen_template_surveys",
    "gen_greedy_surveys",
    "generate_blobs",
    "generate_short_blobs",
    "generate_twilight_near_sun",
)

import copy
from typing import Any

import numpy as np
import numpy.typing as npt

import rubin_scheduler.scheduler.basis_functions as bf
import rubin_scheduler.scheduler.detailers as detailers
from rubin_scheduler.scheduler.surveys import BlobSurvey, GreedySurvey, LongGapSurvey, ScriptedSurvey
from rubin_scheduler.scheduler.utils import ConstantFootprint, Footprints, ecliptic_area
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD

# Set up values to use as kwarg defaults.
NEXP = 1
U_NEXP = 1
EXPTIME = 30.0
U_EXPTIME = 38.0
GOOD_SEEING_MAX = 1.3
CAMERA_ROT_LIMITS = (-80.0, 80.0)
SCIENCE_PROGRAM = "BLOCK-407"

BLOB_SURVEY_PARAMS_DEFAULTS = {
    "slew_approx": 7.5,
    "band_change_approx": 140.0,
    "read_approx": 3.07,
    "flush_time": 30.0,
    "smoothing_kernel": None,
    "nside": DEFAULT_NSIDE,
    "seed": 42,
    "dither": "night",
    "twilight_scale": True,
}


def safety_masks(
    nside: int = DEFAULT_NSIDE,
    moon_distance: float = 30,
    wind_speed_maximum: float = 20.0,
    min_alt: float = 20,
    max_alt: float = 86.5,
    min_az: float = 0,
    max_az: float = 360,
    shadow_minutes: float = 70,
    min_az_sunrise: float = 120,
    max_az_sunrise: float = 290,
    time_to_sunrise: float = 3.0,
    apply_time_limited_shadow: bool = True,
) -> list[bf.BaseBasisFunction]:
    """Basic safety mask basis functions.

    Avoids the moon, bright planets, high wind, and
    areas on the sky out of bounds, using
    the MoonAvoidanceBasisFunction, PlanetMaskBasisFunction,
    AvoidDirectWindBasisFunction, and the AltAzShadowMaskBasisFunction.
    Adds the default AltAzShadowMaskTimeLimited basis function to avoid
    pointing toward sunrise late in the night during commissioning.

    Parameters
    ----------
    nside : `int` or None
        The healpix nside to use.
        Default of None uses rubin_scheduler.utils.get_default_nside.
    moon_distance : `float`, optional
        Moon avoidance distance, in degrees.
    wind_speed_maximum : `float`, optional
        Wind speed maximum to apply to the wind avoidance basis function,
        in m/s.
    min_alt : `float`, optional
        Minimum altitude (in degrees) to observe.
    max_alt : `float`, optional
        Maximum altitude (in degrees) to observe.
    min_az : `float`, optional
        Minimum azimuth angle (in degrees) to observe.
    max_az : `float`, optional
        Maximum azimuth angle (in degrees) to observe.
    shadow_minutes : `float`, optional
        Avoid inaccessible alt/az regions, as well as parts of the sky
        which will move into those regions within `shadow_minutes` (minutes).
        Should be set to the expected time needed to execute the
        observing block associated with the survey.
    min_az_sunrise : `float`, optional
        Minimum azimuth angle (in degrees) to observe during time period
        at the end of the night (during time_to_sunrise).
    max_az_sunrise: `float`, optional
        Maximum azimuth angle (in degrees) to observe during the time period
        at the end of the night (during time_to_sunrise).
    time_to_sunrise : `float`, optional
        Hours before daybreak (sun @ alt=0) to start the azimuth avoidance
        mask.
    apply_time_limited_shadow : `float`, optional
        Flag for whether to apply the morning (time_to_sunrise) azimuth mask.

    Returns
    -------
    mask_basis_functions : `list` [`BaseBasisFunction`]
        Mask basis functions should always be used with a weight of 0.
        The masked (np.nan or -np.inf) regions will remain masked,
        but the basis function values won't influence the reward.
    """
    mask_bfs = []
    # Avoid the moon - too close to the moon will trip the REBs
    mask_bfs.append(bf.MoonAvoidanceBasisFunction(nside=nside, moon_distance=moon_distance))
    # Avoid bright planets
    mask_bfs.append(bf.PlanetMaskBasisFunction(nside=nside))
    # Avoid the wind
    mask_bfs.append(bf.AvoidDirectWind(nside=nside, wind_speed_maximum=wind_speed_maximum))
    # Avoid the alt/az limits - this will pick up limits from the
    # yaml file configurations for the summit as well
    mask_bfs.append(
        bf.AltAzShadowMaskBasisFunction(
            nside=nside,
            min_alt=min_alt,
            max_alt=max_alt,
            min_az=min_az,
            max_az=max_az,
            shadow_minutes=shadow_minutes,
        )
    )

    if apply_time_limited_shadow:
        # Only look away from the azimuth of the sun in the next day
        # permitting emergency dome closure
        mask_bfs.append(
            bf.AltAzShadowTimeLimitedBasisFunction(
                nside=nside,
                min_alt=min_alt,
                max_alt=max_alt,
                min_az=min_az_sunrise,
                max_az=max_az_sunrise,
                shadow_minutes=shadow_minutes,
                # Time until/after sun_keys in hours
                time_to_sun=time_to_sunrise + shadow_minutes / 60.0,
                # 'sunrise' is 0 degree sunrise
                sun_keys=["sunrise"],
            )
        )
    return mask_bfs


def standard_bf(
    nside: int = DEFAULT_NSIDE,
    bandname: str = "g",
    bandname2: str | None = "i",
    m5_weight: float = 6.0,
    footprint_weight: float = 1.5,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    footprints: Footprints | None = None,
    fiducial_fwhm: float = 1.3,
    season: float = 365.25,
    season_start_hour: float = -4.0,
    season_end_hour: float = 2.0,
    strict: bool = True,
    seeing_fwhm_max: float | None = None,
) -> list[tuple[bf.BaseBasisFunction, float]]:
    """Generate the standard basis functions that are shared by blob surveys

    Parameters
    ----------
    nside : `int`
        The HEALpix nside to use. Defaults to DEFAULT_NSIDE
    bandname : `str`
        The band name for the first observation. Default "g".
    bandname2 : `str`
        The band name for the second in the pair (None if unpaired).
        Default "i".
    m5_weight : `float`
        The weight for the 5-sigma depth difference basis function.
    footprint_weight : `float`
        The weight on the survey footprint basis function.
    slewtime_weight : `float`
        The weight on the slewtime basis function.
    stayband_weight : `float`
        The weight on basis function that tries to stay avoid band changes.
    footprints : `rubin_scheduler.scheduler.utils.Footprints` object
        The desired footprints object. Default of None will work, but is likely
        not desirable.
    fiducial_fwhm : `float`
        The fiducial FWHM for the M5Diff Basis function.
    n_obs_template : `dict`
        The number of observations to take every season in each band.
    season : `float`
        The length of season (i.e., how long before templates expire) (days).
        Default 365.25.
    season_start_hour : `float`
        Hour angle limits to use when gathering templates.
        Default -4 (hours)
    season_end_hour : `float`
       Hour angle limits to use when gathering templates.
       Default +2 (hours)
    strict : `bool`
        If False, use BandChangeBasisFunction which rewards visits in the
        same bandpass as currently in-use.
        If True, use a StrictBandBasisFunction which rewards visits in the
        same bandpass as currently in-use, but also rewards visits in
        different filters if the moon rose/set or twilight ended/started,
        or if there was a large gap in observing.
    seeing_fwhm_max : `float`
        Seeing limit to pass to the FootprintBasisFunction - visits
        with delivered image quality > seeing_fwhm_max will not be counted.

    Returns
    -------
    basis_functions_weights : `list`
        list of tuple pairs (basis function, weight) that is
        (rubin_scheduler.scheduler.BasisFunction object, float)

    """

    bfs = []

    if bandname2 is not None:
        bfs.append(
            (
                bf.M5DiffBasisFunction(bandname=bandname, nside=nside, fiducial_FWHMEff=fiducial_fwhm),
                m5_weight / 2.0,
            )
        )
        bfs.append(
            (
                bf.M5DiffBasisFunction(bandname=bandname2, nside=nside, fiducial_FWHMEff=fiducial_fwhm),
                m5_weight / 2.0,
            )
        )

    else:
        bfs.append(
            (
                bf.M5DiffBasisFunction(bandname=bandname, nside=nside, fiducial_FWHMEff=fiducial_fwhm),
                m5_weight,
            )
        )

    if bandname2 is not None:
        bfs.append(
            (
                bf.FootprintBasisFunction(
                    bandname=bandname,
                    footprint=footprints,
                    out_of_bounds_val=np.nan,
                    nside=nside,
                    seeing_fwhm_max=seeing_fwhm_max,
                ),
                footprint_weight / 2.0,
            )
        )
        bfs.append(
            (
                bf.FootprintBasisFunction(
                    bandname=bandname2,
                    footprint=footprints,
                    out_of_bounds_val=np.nan,
                    nside=nside,
                    seeing_fwhm_max=seeing_fwhm_max,
                ),
                footprint_weight / 2.0,
            )
        )
    else:
        bfs.append(
            (
                bf.FootprintBasisFunction(
                    bandname=bandname,
                    footprint=footprints,
                    out_of_bounds_val=np.nan,
                    nside=nside,
                    seeing_fwhm_max=seeing_fwhm_max,
                ),
                footprint_weight,
            )
        )

    bfs.append(
        (
            bf.SlewtimeBasisFunction(bandname=bandname, nside=nside),
            slewtime_weight,
        )
    )
    if strict:
        bfs.append((bf.StrictBandBasisFunction(bandname=bandname), stayband_weight))
    else:
        bfs.append((bf.BandChangeBasisFunction(bandname=bandname), stayband_weight))

    bandnames = [fn for fn in [bandname, bandname2] if fn is not None]
    bfs.append((bf.BandLoadedBasisFunction(bandnames=bandnames), 0))

    return bfs


def gen_template_surveys(
    footprints: Footprints,
    nside: int = DEFAULT_NSIDE,
    seeing_fwhm_max: float = GOOD_SEEING_MAX,
    band1s: list[str] = ["u", "g", "r", "i", "z", "y"],
    dark_only: list[str] = ["u", "g"],
    ignore_obs: str | list[str] = ["DD", "twilight_near_sun"],
    camera_rot_limits: tuple[float, float] = CAMERA_ROT_LIMITS,
    exptime: float = EXPTIME,
    nexp: int = NEXP,
    u_exptime: float = U_EXPTIME,
    u_nexp: int = U_NEXP,
    n_obs_template: dict | None = None,
    pair_time: float = 33.0,
    area_required: float = 50.0,
    HA_min: float = 2.5,
    HA_max: float = 24 - 2.5,
    max_alt: float = 76.0,
    m5_weight: float = 6.0,
    footprint_weight: float = 1.5,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    science_program: str = SCIENCE_PROGRAM,
    blob_survey_params: dict | None = None,
    safety_mask_params: dict | None = None,
) -> list[BlobSurvey]:
    """Coherent area surveys (BlobSurvey with single visit) that
    are intended to acquire template visits in a convenient yet
    aggressive manner.

    Parameters
    ----------
    footprints : `rubin_scheduler.scheduler.utils.Footprints`
        The Footprints object for the Surveys.
    nside : `int`
        Nside for the surveys.
    seeing_fwhm_max : `float`
        The maximum (model delivered) seeing acceptable for the template
        surveys. This limit is set to attempt to match the seeing limits
        used when building difference image coadds.
    band1s : `list` [ `str` ]
        The bands in which to obtain templates, within the Footprints.
    dark_only : `list` [ `str` ]
        The bands to only attempt during dark-time.
    ignore_obs : `str` or `list` [ `str` ]
        Strings to match within scheduler_note to flag observations to ignore.
    camera_rot_limits : `list` [ `float`, `float` ]
        Camera rotator limits (in degrees) for the dither rotation detailer.
    exptime : `float`
        The exposure time for grizy visits.
    nexp : `int`
        The number of exposures per visit for grizy visits.
    u_exptime : `float`
        The exposure time for u band visits.
    u_nexp : `int`
        The number of exposures per visit for u band visits.
    n_obs_template : `dict` { `str` : `int` }
        Number of visits per bandpass before masking the Survey healpix.
        When this many visits are acquired, the pixel is considered "done".
    pair_time : `float`
        The time until the end of the first pass of the Blob.
        Since there is no second filter, this is the amount of time
        spent in the Blob.
    area_required : `float`
        The area required that needs templates, before the BlobSurvey will
        activate. Square degrees.
    HA_min : `float`
        The minimum HA to consider when considering template area.
    HA_max : `float`
        The maximum HA to consider when considering template area.
    max_alt : `float`
        The maximum altitude to use for the Surveys.
        Typically for BlobSurveys this is set lower than the max available,
        to about 76 degrees, to avoid long dome slews near azimuth.
        This is masked separately from the `safety_masks`.
    m5_weight : `float`
        The weight for the 5-sigma depth difference basis function.
    footprint_weight : `float`
        The weight on the survey footprint basis function.
    slewtime_weight : `float`
        The weight on the slewtime basis function.
    stayband_weight : `float`
        The weight on basis function that tries to stay avoid band changes.
    science_program : `str`
        The science_program to use for visits from these surveys.
    blob_survey_params : `dict` or None
        A dictionary of additional kwargs to pass to the BlobSurvey.
        In particular, the times for typical slews, readtime, etc. are
        useful for setting the number of pointings to schedule within
        pair_time.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.
    """

    if n_obs_template is None:
        n_obs_template = {"u": 4, "g": 4, "r": 4, "i": 4, "z": 4, "y": 4}

    if blob_survey_params is None:
        blob_survey_params = BLOB_SURVEY_PARAMS_DEFAULTS

    if safety_mask_params is None:
        safety_mask_params = {}
        safety_mask_params["nside"] = nside
    else:
        safety_mask_params = copy.deepcopy(safety_mask_params)
    if "shadow_minutes" not in safety_mask_params or safety_mask_params["shadow_minutes"] < pair_time:
        # Shadow minutes is only pair_time, only taking 1 image per field
        safety_mask_params["shadow_minutes"] = pair_time

    surveys = []

    for bandname in band1s:
        # Set up Detailers for camera rotator and ordering in altitude.
        detailer_list = []
        detailer_list.append(
            detailers.CameraRotDetailer(min_rot=np.min(camera_rot_limits), max_rot=np.max(camera_rot_limits))
        )
        detailer_list.append(detailers.CloseAltDetailer())
        detailer_list.append(detailers.BandNexp(bandname="u", nexp=u_nexp, exptime=u_exptime))
        detailer_list.append(detailers.LabelRegionsAndDDFs())

        # List to hold tuples of (basis_function_object, weight)
        bfs = []

        bfs.extend(
            standard_bf(
                nside,
                bandname=bandname,
                bandname2=None,
                footprints=footprints,
                m5_weight=m5_weight,
                footprint_weight=footprint_weight,
                slewtime_weight=slewtime_weight,
                stayband_weight=stayband_weight,
                seeing_fwhm_max=seeing_fwhm_max,
            )
        )

        bfs.append(
            (
                bf.AltAzShadowMaskBasisFunction(
                    nside=nside, shadow_minutes=pair_time, max_alt=max_alt, pad=3.0
                ),
                0,
            )
        )

        # not in twilight bf
        bfs.append((bf.TimeToTwilightBasisFunction(time_needed=pair_time), 0.0))
        bfs.append((bf.NotTwilightBasisFunction(), 0.0))
        bfs.append((bf.RevHaMaskBasisFunction(ha_min=HA_min, ha_max=HA_max, nside=nside), 0.0))

        # Need a once in night mask
        bfs.append((bf.NInNightMaskBasisFunction(n_limit=1, nside=nside), 0.0))

        # If u or g, only when moon is down
        if bandname in dark_only:
            bfs.append((bf.MoonAltLimitBasisFunction(alt_limit=-5), 0.0))

        # limit to first year
        bfs.append((bf.OnlyBeforeNightBasisFunction(night_max=366), 0.0))

        # Limit to only good seeing visits.
        bfs.append(
            (
                bf.MaskPoorSeeing(bandname, nside=nside, seeing_fwhm_max=seeing_fwhm_max),
                0,
            )
        )

        # Mask anything observed n_obs_template times reseting each season
        bfs.append(
            (
                bf.MaskAfterNObsSeeingBasisFunction(
                    nside=nside,
                    n_max=n_obs_template[bandname],
                    bandname=bandname,
                    seeing_fwhm_max=seeing_fwhm_max,
                ),
                0.0,
            )
        )

        # Add safety masks
        masks = safety_masks(**safety_mask_params)
        for m in masks:
            bfs.append((m, 0))

        # unpack the basis functions and weights
        weights = [val[1] for val in bfs]
        basis_functions = [val[0] for val in bfs]

        survey_name = "templates, %s" % bandname

        surveys.append(
            BlobSurvey(
                basis_functions,
                weights,
                bandname1=bandname,
                bandname2=None,
                exptime=exptime,
                ideal_pair_time=pair_time,
                survey_name=survey_name,
                science_program=science_program,
                observation_reason=f"template_blob_{bandname}_{pair_time:.1f}",
                ignore_obs=ignore_obs,
                nexp=nexp,
                detailers=detailer_list,
                area_required=area_required,
                **blob_survey_params,
            )
        )

    return surveys


def blob_for_long(
    footprints: Footprints,
    nside: int = DEFAULT_NSIDE,
    band1s: list[str] = ["g"],
    band2s: list[str] = ["i"],
    ignore_obs: str | list[str] = ["DD", "twilight_near_sun", "ToO"],
    camera_rot_limits: tuple[float, float] = CAMERA_ROT_LIMITS,
    exptime: float = EXPTIME,
    nexp: int = NEXP,
    u_exptime: float = U_EXPTIME,
    u_nexp: int = U_NEXP,
    n_obs_template: dict | None = None,
    pair_time: float = 33.0,
    season: float = 365.25,
    season_start_hour: float = -4.0,
    season_end_hour: float = 2.0,
    HA_min: float = 12,
    HA_max: float = 24 - 3.5,
    max_alt: float = 76.0,
    m5_weight: float = 6.0,
    footprint_weight: float = 1.5,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    night_pattern: list[bool] = [True, True],
    time_after_twi: float = 30.0,
    blob_names: list[str] = [],
    scheduled_respect: float = 30.0,
    science_program: str = SCIENCE_PROGRAM,
    observation_reason: str | None = None,
    blob_survey_params: dict | None = None,
    safety_mask_params: dict | None = None,
    pair_pad: float = 5.0,
) -> list[BlobSurvey]:
    """
    Generate surveys that take observations in blobs.

    Parameters
    ----------
    footprints : `rubin_scheduler.scheduler.utils.Footprints`
        The Footprints object for the Surveys.
    nside : `int`
        The HEALpix nside to use. Default to DEFAULT_NSIDE.
    band1s : `list` [`str`]
        The bandnames for the first band in a pair.
        Default ["g"].
    band2s : `list` of `str`
        The band names for the second in the pair (None if unpaired).
        Default ["i"].
    ignore_obs : `str` or `list` of `str`
        Ignore observations by surveys that include the given substring(s).
    camera_rot_limits : `list` of `float`
        The limits to impose when rotationally dithering the camera (degrees).
        Default [-80., 80.].
    exptime : `float`
        The exposure time for grizy visits.
    nexp : `int`
        The number of exposures per visit for grizy visits.
    u_exptime : `float`
        The exposure time for u band visits.
    u_nexp : `int`
        The number of exposures per visit for u band visits.
    n_obs_template : `dict`
        The number of observations to take every season in each band.
        If None, sets to 3 each. Default None.
    pair_time : `float`
        The ideal time between pairs (minutes). Default 33.
    season : float
        The length of season (i.e., how long before templates expire) (days)
        Default 365.25.
    season_start_hour : `float`
        Hour angle limits to use when gathering templates.
        Default -4 (hours)
    sesason_end_hour : `float`
       Hour angle limits to use when gathering templates.
       Default +2 (hours)
    HA_min : `float`
        The minimum HA to consider when considering template area.
    HA_max : `float`
        The maximum HA to consider when considering template area.
    max_alt : `float`
        The maximium altitude to use when masking zenith (degrees).
        Typically for BlobSurveys this is set lower than the max available,
        to about 76 degrees, to avoid long dome slews near azimuth.
        This is masked separately from the `safety_masks`.
        Default 76.
    m5_weight : `float`
        The weight for the 5-sigma depth difference basis function.
    footprint_weight : `float`
        The weight on the survey footprint basis function.
    slewtime_weight : `float`
        The weight on the slewtime basis function.
    stayband_weight : `float`
        The weight on basis function that tries to stay avoid band changes.
    night_pattern : `list` [ `bool` ]
        Flips the blob_for_long surveys on or off, on a pattern of
        nights. These surveys typically don't execute every night.
    time_after_twi : `float`
        Don't start before this many minutes pass after -18 degree twilight.
    blob_names : `list` [ `str` ]
        Strings to match in scheduler_note to ensure surveys don't execute
        more than once per night.
    scheduled_respect : `float`
        Ensure that blobs don't start within this many minutes of scheduled
        observations (from a ScriptedSurvey).
    blob_survey_params : `dict` or None
        A dictionary of additional kwargs to pass to the BlobSurvey.
        In particular, the times for typical slews, readtime, etc. are
        useful for setting the number of pointings to schedule within
        pair_time.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.
    pair_pad : `float`
        How much extra time to pad above the necessary pair time
        for shadow basis function.
    """

    if blob_survey_params is None:
        blob_survey_params = BLOB_SURVEY_PARAMS_DEFAULTS
    if safety_mask_params is None:
        safety_mask_params = {"nside": nside}
    else:
        safety_mask_params = copy.deepcopy(safety_mask_params)

    shadow_minutes = pair_time * 2 + pair_pad
    if "shadow_minutes" not in safety_mask_params or safety_mask_params["shadow_minutes"] < shadow_minutes:
        safety_mask_params["shadow_minutes"] = shadow_minutes

    surveys = []
    if n_obs_template is None:
        n_obs_template = {"u": 3, "g": 3, "r": 3, "i": 3, "z": 3, "y": 3}

    times_needed = [pair_time, pair_time * 2]
    for bandname, bandname2 in zip(band1s, band2s):
        detailer_list = []
        detailer_list.append(
            detailers.CameraRotDetailer(min_rot=np.min(camera_rot_limits), max_rot=np.max(camera_rot_limits))
        )
        detailer_list.append(detailers.CloseAltDetailer())
        detailer_list.append(detailers.BandNexp(bandname="u", nexp=u_nexp, exptime=u_exptime))
        detailer_list.append(detailers.LabelRegionsAndDDFs())

        # List to hold tuples of (basis_function_object, weight)
        bfs = []

        bfs.extend(
            standard_bf(
                nside,
                bandname=bandname,
                bandname2=bandname2,
                m5_weight=m5_weight,
                footprint_weight=footprint_weight,
                slewtime_weight=slewtime_weight,
                stayband_weight=stayband_weight,
                footprints=footprints,
                season=season,
                season_start_hour=season_start_hour,
                season_end_hour=season_end_hour,
            )
        )

        # Make sure we respect scheduled observations
        bfs.append((bf.TimeToScheduledBasisFunction(time_needed=scheduled_respect), 0))

        bfs.append(
            (
                bf.AltAzShadowMaskBasisFunction(
                    nside=nside, shadow_minutes=shadow_minutes, max_alt=max_alt, pad=3.0
                ),
                0.0,
            )
        )
        if bandname2 is None:
            time_needed = times_needed[0]
        else:
            time_needed = times_needed[1]
        bfs.append((bf.TimeToTwilightBasisFunction(time_needed=time_needed), 0.0))
        bfs.append((bf.NotTwilightBasisFunction(), 0.0))
        bfs.append((bf.AfterEveningTwiBasisFunction(time_after=time_after_twi), 0.0))
        bfs.append((bf.HaMaskBasisFunction(ha_min=HA_min, ha_max=HA_max, nside=nside), 0.0))
        # don't execute every night
        bfs.append((bf.NightModuloBasisFunction(night_pattern), 0.0))
        # only execute one blob per night
        bfs.append((bf.OnceInNightBasisFunction(notes=blob_names), 0))

        # Add safety masks
        masks = safety_masks(**safety_mask_params)
        for m in masks:
            bfs.append((m, 0))

        # unpack the basis functions and weights
        weights = [val[1] for val in bfs]
        basis_functions = [val[0] for val in bfs]
        if bandname2 is None:
            survey_name = "blob_long, %s" % bandname
        else:
            survey_name = "blob_long, %s%s" % (bandname, bandname2)
        if bandname2 is not None:
            detailer_list.append(detailers.TakeAsPairsDetailer(bandname=bandname2))

        if observation_reason is None:
            observation_reason = f"triplet_pairs_{bandname}{bandname2}_{pair_time:.1f}"

        surveys.append(
            BlobSurvey(
                basis_functions,
                weights,
                bandname1=bandname,
                bandname2=bandname2,
                exptime=exptime,
                ideal_pair_time=pair_time,
                survey_name=survey_name,
                ignore_obs=ignore_obs,
                nexp=nexp,
                detailers=detailer_list,
                science_program=science_program,
                observation_reason=observation_reason,
                **blob_survey_params,
            )
        )

    return surveys


def gen_long_gaps_survey(
    footprints: Footprints,
    nside: int = DEFAULT_NSIDE,
    camera_rot_limits: tuple[float, float] = CAMERA_ROT_LIMITS,
    exptime: float = EXPTIME,
    nexp: int = NEXP,
    u_exptime: float = U_EXPTIME,
    u_nexp: int = U_NEXP,
    pair_time: float = 33.0,
    night_pattern: list[bool] = [True, True],
    gap_range: list[float] = [2, 7],
    HA_min: float = 12,
    HA_max: float = 24 - 3.5,
    time_after_twi: float = 120,
    m5_weight: float = 6.0,
    footprint_weight: float = 1.5,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    science_program: str = SCIENCE_PROGRAM,
    blob_survey_params: dict | None = None,
    safety_mask_params: dict | None = None,
    pair_pad: float = 5,
) -> list[LongGapSurvey]:
    """Generate long-gaps (triplets) surveys.

    Parameters
    -----------
    footprints : `rubin_scheduler.scheduler.utils.footprints.Footprints`
        The footprints to be used for the long-gaps surveys.
    nside : `int`
        The nside for the surveys.
    camera_rot_limits : `list` of `float`
        The limits to impose when rotationally dithering the camera (degrees).
        Default [-80., 80.].
    exptime : `float`
        The exposure time for grizy visits.
    nexp : `int`
        The number of exposures per visit for grizy visits.
    u_exptime : `float`
        The exposure time for u band visits.
    u_nexp : `int`
        The number of exposures per visit for u band visits.
    pair_time : `float`
        The ideal time between pairs (minutes). Default 33.
    night_pattern : `list` [ `bool` ]
        Which nights to let the survey execute.
        Default of [True, True] executes every night.
    gap_range : `list` [ `float` ]
        Range of times to attempt to gather pairs (hours).
        Default [2, 7].
    HA_min : `float`
        The hour angle limits passed to the initial blob scheduler.
        Default 12 (hours)
    HA_max : `float`
        The hour angle limits passed to the initial blob scheduler.
        Default 20.5 (hours).
    time_after_twi : `float`
        The time after evening twilight to attempt long gaps (minutes).
        Default 120.
    m5_weight : `float`
        The weight for the 5-sigma depth difference basis function.
    footprint_weight : `float`
        The weight on the survey footprint basis function.
    slewtime_weight : `float`
        The weight on the slewtime basis function.
    stayband_weight : `float`
        The weight on basis function that tries to stay avoid band changes.
    science_program : `str`
        The science_program to use for visits from these surveys.
    blob_survey_params : `dict` or None
        A dictionary of additional kwargs to pass to the BlobSurvey.
        In particular, the times for typical slews, readtime, etc. are
        useful for setting the number of pointings to schedule within
        pair_time.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.
    pair_pad : `float`
        How much extra time to pad above the necessary pair time
        for shadow basis function.
    """
    # Only copy the safety_mask_params here for the ScriptedSurvey.
    # The blob_for_long survey will copy/update on its own.
    safety_mask_params_scripted: dict[str, Any] = {"nside": nside}
    if safety_mask_params is not None:
        for key in safety_mask_params:
            safety_mask_params_scripted[key] = safety_mask_params[key]
            if key == "shadow_minutes":
                if safety_mask_params[key] < pair_time:
                    safety_mask_params_scripted[key] = pair_time
    if "shadow_minutes" not in safety_mask_params_scripted:
        safety_mask_params_scripted["shadow_minutes"] = pair_time

    surveys = []
    f1 = ["g", "r", "i"]
    f2 = ["r", "i", "z"]
    # Maybe force scripted to not go in twilight?
    blob_names = []
    for fn1, fn2 in zip(f1, f2):
        for ab in ["a", "b"]:
            blob_names.append("blob_long, %s%s, %s" % (fn1, fn2, ab))
    for bandname1, bandname2 in zip(f1, f2):
        blob = blob_for_long(
            footprints=footprints,
            camera_rot_limits=camera_rot_limits,
            exptime=exptime,
            nexp=nexp,
            u_exptime=u_exptime,
            u_nexp=u_nexp,
            pair_time=pair_time,
            nside=nside,
            band1s=[bandname1],
            band2s=[bandname2],
            night_pattern=night_pattern,
            time_after_twi=time_after_twi,
            HA_min=HA_min,
            HA_max=HA_max,
            m5_weight=m5_weight,
            footprint_weight=footprint_weight,
            slewtime_weight=slewtime_weight,
            stayband_weight=stayband_weight,
            blob_names=blob_names,
            science_program=science_program,
            blob_survey_params=blob_survey_params,
            safety_mask_params=safety_mask_params,
            pair_pad=pair_pad,
        )
        masks = safety_masks(**safety_mask_params_scripted)
        scripted = ScriptedSurvey(
            masks,
            nside=nside,
            ignore_obs=["blob", "DDF", "twi", "pair", "templates", "ToO"],
            science_program=science_program,
            detailers=[detailers.LabelRegionsAndDDFs()],
        )
        surveys.append(LongGapSurvey(blob[0], scripted, gap_range=gap_range, avoid_zenith=True))

    return surveys


def gen_greedy_surveys(
    nside: int = DEFAULT_NSIDE,
    bands: list[str] = ["r", "i", "z", "y"],
    ignore_obs: list[str] = ["DD", "twilight_near_sun", "ToO"],
    camera_rot_limits: tuple[float, float] = CAMERA_ROT_LIMITS,
    exptime: float = EXPTIME,
    nexp: int = NEXP,
    u_exptime: float = U_EXPTIME,
    u_nexp: int = U_NEXP,
    shadow_minutes: float = 15.0,
    max_alt: float = 76.0,
    m5_weight: float = 3.0,
    footprint_weight: float = 0.75,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 100.0,
    repeat_weight: float = -1.0,
    footprints: Footprints | None = None,
    science_program: str = SCIENCE_PROGRAM,
    safety_mask_params: dict | None = None,
) -> list[GreedySurvey]:
    """Generate greedy (single-best choice visits) Surveys.

    Parameters
    ----------
    nside : `int`
        The HEALpix nside to use
    bands : `list` [ `str` ]
        Bands in which to generate greedy surveys.
        Default ['r', 'i', 'z', 'y'].
    ignore_obs : `str` or `list` of `str`
        Ignore observations by surveys that include the given substring(s).
    camera_rot_limits : `list` [ `float` ]
        The limits to impose when rotationally dithering the camera (degrees).
        Default [-80., 80.].
    exptime : `float`
        The exposure time for grizy visits.
    nexp : `int`
        The number of exposures per visit for grizy visits.
    u_exptime : `float`
        The exposure time for u band visits.
    u_nexp : `int`
        The number of exposures per visit for u band visits.
    shadow_minutes : `float`
        Used to mask regions around zenith (minutes).
    max_alt : `float`
        The maximium altitude to use when masking zenith (degrees).
    m5_weight : `float`
        The weight for the 5-sigma depth difference basis function.
    footprint_weight : `float`
        The weight on the survey footprint basis function.
    slewtime_weight : `float`
        The weight on the slewtime basis function.
    stayband_weight : `float`
        The weight on basis function that tries to stay avoid band changes.
    repeat_weight : `float`
        Weight that enhances (+ values) or decreases (- values) the likelihood
        of revisiting the same pointing within a two-hour time gap.
    footprints : `rubin_scheduler.scheduler.utils.footprints.Footprints`
        The footprints to be used for the long-gaps surveys.
    science_program : `str`
        The science_program to use for visits from these surveys.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.
    """
    # Define the extra parameters that are used in the greedy survey. I
    # think these are fairly set, so no need to promote to utility func kwargs
    greed_survey_params = {
        "block_size": 1,
        "smoothing_kernel": None,
        "seed": 42,
        "camera": "LSST",
        "dither": "night",
    }
    if safety_mask_params is None:
        safety_mask_params = {"nside": nside}
    else:
        safety_mask_params = copy.deepcopy(safety_mask_params)
    if "shadow_minutes" not in safety_mask_params or safety_mask_params["shadow_minutes"] < shadow_minutes:
        safety_mask_params["shadow_minutes"] = shadow_minutes

    surveys = []
    detailer_list = [
        detailers.CameraRotDetailer(min_rot=np.min(camera_rot_limits), max_rot=np.max(camera_rot_limits))
    ]
    detailer_list.append(detailers.LabelRegionsAndDDFs())

    if "u" in bands:
        detailer_list.append(detailers.BandNexp(bandname="u", nexp=u_nexp, exptime=u_exptime))

    for bandname in bands:
        bfs = []
        bfs.extend(
            standard_bf(
                nside,
                bandname=bandname,
                bandname2=None,
                m5_weight=m5_weight,
                footprint_weight=footprint_weight,
                slewtime_weight=slewtime_weight,
                stayband_weight=stayband_weight,
                footprints=footprints,
                strict=False,
            )
        )

        # XXX-magic numbers
        bfs.append(
            (
                bf.VisitRepeatBasisFunction(
                    gap_min=0, gap_max=2 * 60.0, bandname=None, nside=nside, npairs=20
                ),
                repeat_weight,
            )
        )
        # Masks, give these 0 weight
        bfs.append(
            (
                bf.AltAzShadowMaskBasisFunction(
                    nside=nside, shadow_minutes=shadow_minutes, max_alt=max_alt, pad=3.0
                ),
                0,
            )
        )

        masks = safety_masks(nside=nside, shadow_minutes=shadow_minutes)
        for m in masks:
            bfs.append((m, 0))

        weights = [val[1] for val in bfs]
        basis_functions = [val[0] for val in bfs]
        surveys.append(
            GreedySurvey(
                basis_functions,
                weights,
                exptime=exptime,
                bandname=bandname,
                nside=nside,
                ignore_obs=ignore_obs,
                nexp=nexp,
                detailers=detailer_list,
                survey_name=f"greedy {bandname}",
                science_program=science_program,
                observation_reason=f"singles_{bandname}",
                **greed_survey_params,
            )
        )

    return surveys


def generate_blobs(
    footprints: Footprints,
    nside: int = DEFAULT_NSIDE,
    band1s: list[str] = ["u", "u", "g", "r", "i", "z", "y"],
    band2s: list[str] = ["g", "r", "r", "i", "z", "y", "y"],
    ignore_obs: str | list[str] = ["DD", "twilight_near_sun", "ToO"],
    camera_rot_limits: tuple[float, float] = CAMERA_ROT_LIMITS,
    exptime: float = EXPTIME,
    nexp: int = NEXP,
    u_exptime: float = U_EXPTIME,
    u_nexp: int = U_NEXP,
    n_obs_template: dict | None = None,
    pair_time: float = 33.0,
    season: float = 365.25,
    season_start_hour: float = -4.0,
    season_end_hour: float = 2.0,
    max_alt: float = 76.0,
    seeing_fwhm_best: float = 0.8,
    m5_penalty_max: float = 0.5,
    m5_weight: float = 6.0,
    footprint_weight: float = 1.5,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    repeat_weight: float = -20,
    good_seeing: dict = {"g": 3, "r": 3, "i": 3},
    good_seeing_weight: float = 3.0,
    survey_start: float = SURVEY_START_MJD,
    scheduled_respect: float = 45.0,
    science_program: str = SCIENCE_PROGRAM,
    blob_survey_params: dict | None = None,
    safety_mask_params: dict | None = None,
    pair_pad: float = 5.0,
) -> list[BlobSurvey]:
    """Generate surveys that take observations in blobs.

    Parameters
    ----------
    footprints : `rubin_scheduler.scheduler.utils.Footprints`
        The Footprints object for the Surveys.
    nside : `int`
        The HEALpix nside to use. Default to DEFAULT_NSIDE.
    band1s : `list` [`str`]
        The bandnames for the first band in a pair.
    band2s : `list` of `str`
        The band names for the second in the pair (None if unpaired).
    ignore_obs : `str` or `list` of `str`
        Ignore observations by surveys that include the given substring(s).
    camera_rot_limits : `list` of `float`
        The limits to impose when rotationally dithering the camera (degrees).
    exptime : `float`
        The exposure time for grizy visits.
    nexp : `int`
        The number of exposures per visit for grizy visits.
    u_exptime : `float`
        The exposure time for u band visits.
    u_nexp : `int`
        The number of exposures per visit for u band visits.
    n_obs_template : `dict`
        The number of observations to take every season in each band.
        If None, sets to 3 each. Default None.
    pair_time : `float`
        The ideal time between pairs (minutes). Default 33.
    season : float
        The length of season (i.e., how long before templates expire) (days)
    season_start_hour : `float`
        Hour angle limits to use when gathering templates.
        Default -4 (hours)
    season_end_hour : `float`
       Hour angle limits to use when gathering templates.
       Default +2 (hours)
    max_alt : `float`
        The maximum altitude to use for the Surveys.
        Typically for BlobSurveys this is set lower than the max available,
        to about 76 degrees, to avoid long dome slews near azimuth.
        This is masked separately from the `safety_masks`.
    seeing_fwhm_best : `float`
        The FWHM effective to use when counting visits for the "GoodSeeing"
        basis function (not for templates; for deblending and shapes).
    m5_penalty_max : `float`
        The maximum penalty in 5-sigma limiting depth to consider
        still good for template building. Default 0.5 (mags).
    m5_weight : `float`
        The weight for the 5-sigma depth difference basis function.
    footprint_weight : `float`
        The weight on the survey footprint basis function.
    slewtime_weight : `float`
        The weight on the slewtime basis function.
    stayband_weight : `float`
        The weight on basis function that tries to stay avoid band changes.
    repeat_weight : `float`
        Weight that enhances (+ values) or decreases (- values) the likelihood
        of revisiting the same pointing within a two-hour time gap.
    good_seeing : `dict`
        Number of good-seeing (<6") images per band for basis function
        directing good-atmospheric seeing into particular bands.
    good_seeing_weight : `float`
        The weight to place on getting good-seeing images in the
        good_seeing bands.
    survey_start : `float`
        The mjd that the survey started (used for determining season for
        counting good seeing images within a season).
    scheduled_respect : `float`
        Ensure that blobs don't start within this many minutes of scheduled
        observations (from a ScriptedSurvey).
    science_program : `str`
        The science_program to use for visits from these surveys.
    blob_survey_params : `dict` or None
        A dictionary of additional kwargs to pass to the BlobSurvey.
        In particular, the times for typical slews, readtime, etc. are
        useful for setting the number of pointings to schedule within
        pair_time.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.
    pair_pad : `float`
        Padding to add to shadow mask. Default 5 (minutes).
    """

    if blob_survey_params is None:
        blob_survey_params = BLOB_SURVEY_PARAMS_DEFAULTS
    if safety_mask_params is None:
        safety_mask_params = {"nside": nside}
    else:
        safety_mask_params = copy.deepcopy(safety_mask_params)

    shadow_minutes = pair_time * 2 + pair_pad
    if "shadow_minutes" not in safety_mask_params or safety_mask_params["shadow_minutes"] < shadow_minutes:
        safety_mask_params["shadow_minutes"] = shadow_minutes
    if "max_alt" in safety_mask_params and safety_mask_params["max_alt"] > max_alt:
        safety_mask_params["max_alt"] = max_alt

    if n_obs_template is None:
        n_obs_template = {"u": 3, "g": 3, "r": 3, "i": 3, "z": 3, "y": 3}

    surveys = []

    times_needed = [pair_time, pair_time * 2]
    for bandname, bandname2 in zip(band1s, band2s):
        detailer_list = []
        detailer_list.append(
            detailers.CameraRotDetailer(min_rot=np.min(camera_rot_limits), max_rot=np.max(camera_rot_limits))
        )
        detailer_list.append(detailers.CloseAltDetailer())
        if (bandname == "u") | (bandname2 == "u"):
            detailer_list.append(detailers.BandNexp(bandname="u", nexp=U_NEXP, exptime=u_exptime))
        detailer_list.append(detailers.FlushForSchedDetailer())
        detailer_list.append(detailers.LabelRegionsAndDDFs())
        # List to hold tuples of (basis_function_object, weight)
        bfs = []

        bfs.extend(
            standard_bf(
                nside,
                bandname=bandname,
                bandname2=bandname2,
                m5_weight=m5_weight,
                footprint_weight=footprint_weight,
                slewtime_weight=slewtime_weight,
                stayband_weight=stayband_weight,
                footprints=footprints,
                season=season,
                season_start_hour=season_start_hour,
                season_end_hour=season_end_hour,
            )
        )

        # Suppress revisits within 3 hours of the first pair.
        # Without this, we tend to repeat fields too quickly
        # but repeats after 3 hours could be useful, so let it expire then.
        bfs.append(
            (
                bf.VisitRepeatBasisFunction(
                    gap_min=0, gap_max=3 * 60.0, bandname=None, nside=nside, npairs=20
                ),
                repeat_weight,
            )
        )

        # Insert things for getting good seeing images
        # Probably for galaxy shape measurements.
        if bandname2 is not None:
            if bandname in list(good_seeing.keys()):
                bfs.append(
                    (
                        bf.NGoodSeeingBasisFunction(
                            bandname=bandname,
                            nside=nside,
                            mjd_start=survey_start,
                            footprint=footprints.get_footprint(bandname),
                            n_obs_desired=good_seeing[bandname],
                            seeing_fwhm_max=seeing_fwhm_best,
                            m5_penalty_max=m5_penalty_max,
                        ),
                        good_seeing_weight,
                    )
                )
            if bandname2 in list(good_seeing.keys()):
                bfs.append(
                    (
                        bf.NGoodSeeingBasisFunction(
                            bandname=bandname2,
                            nside=nside,
                            mjd_start=survey_start,
                            footprint=footprints.get_footprint(bandname2),
                            n_obs_desired=good_seeing[bandname2],
                            seeing_fwhm_max=seeing_fwhm_best,
                            m5_penalty_max=m5_penalty_max,
                        ),
                        good_seeing_weight,
                    )
                )
        else:
            if bandname in list(good_seeing.keys()):
                bfs.append(
                    (
                        bf.NGoodSeeingBasisFunction(
                            bandname=bandname,
                            nside=nside,
                            mjd_start=survey_start,
                            footprint=footprints.get_footprint(bandname),
                            n_obs_desired=good_seeing[bandname],
                            seeing_fwhm_max=seeing_fwhm_best,
                            m5_penalty_max=m5_penalty_max,
                        ),
                        good_seeing_weight,
                    )
                )
        # Make sure we respect scheduled observations
        bfs.append((bf.TimeToScheduledBasisFunction(time_needed=scheduled_respect), 0))

        if bandname2 is None:
            time_needed = times_needed[0]
        else:
            time_needed = times_needed[1]
        bfs.append((bf.TimeToTwilightBasisFunction(time_needed=time_needed), 0.0))
        bfs.append((bf.NotTwilightBasisFunction(), 0.0))

        # Add safety masks
        masks = safety_masks(**safety_mask_params)
        for m in masks:
            bfs.append((m, 0))

        # unpack the basis functions and weights
        weights = [val[1] for val in bfs]
        basis_functions = [val[0] for val in bfs]

        # Set survey name
        if bandname2 is None:
            survey_name = "pair_%i, %s" % (pair_time, bandname)
        else:
            survey_name = "pair_%i, %s%s" % (pair_time, bandname, bandname2)
        if bandname2 is not None:
            detailer_list.append(detailers.TakeAsPairsDetailer(bandname=bandname2))

        observation_reason = f"pairs_{bandname}"
        if bandname2 is not None:
            observation_reason += f"{bandname2}"
        observation_reason += f"_{pair_time:.1f}"

        surveys.append(
            BlobSurvey(
                basis_functions,
                weights,
                bandname1=bandname,
                bandname2=bandname2,
                exptime=exptime,
                ideal_pair_time=pair_time,
                survey_name=survey_name,
                science_program=science_program,
                observation_reason=observation_reason,
                ignore_obs=ignore_obs,
                nexp=nexp,
                detailers=detailer_list,
                **blob_survey_params,
            )
        )

    return surveys


def generate_short_blobs(
    footprints: Footprints,
    nside: int = DEFAULT_NSIDE,
    band1s: list[str] = ["r", "i", "z", "y"],
    band2s: list[str] = ["i", "z", "y", "y"],
    ignore_obs: str | list[str] = ["DD", "twilight_near_sun", "ToO"],
    camera_rot_limits: tuple[float, float] = CAMERA_ROT_LIMITS,
    exptime: float = EXPTIME,
    nexp: int = NEXP,
    n_obs_template: dict[str, int] | None = None,
    pair_time: float = 15.0,
    season: float = 365.25,
    season_start_hour: float = -4.0,
    season_end_hour: float = 2.0,
    max_alt: float = 76.0,
    m5_weight: float = 6.0,
    footprint_weight: float = 1.5,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    repeat_weight: float = 0,
    scheduled_respect: float = 15.0,
    night_pattern: list[bool] | None = None,
    science_program: str = SCIENCE_PROGRAM,
    blob_survey_params: dict | None = None,
    safety_mask_params: dict | None = None,
    pair_pad: float = 5.0,
) -> list[BlobSurvey]:
    """
    Generate surveys that take observations in blobs, for twilight time.
    Shorter blobs, different weights for the basis functions.

    Parameters
    ----------
    footprints : `rubin_scheduler.scheduler.utils.Footprints`
        The Footprints object for the Surveys.
    nside : `int`
        The HEALpix nside to use. Default to DEFAULT_NSIDE.
    band1s : `list` [`str`]
        The bandnames for the first band in a pair.
    band2s : `list` [ `str` ]
        The band names for the second in the pair (None if unpaired).
    ignore_obs : `str` or `list` [ `str` ]
        Ignore observations by surveys that include the given substring(s).
    camera_rot_limits : `list` [ `float` ]
        The limits to impose when rotationally dithering the camera (degrees).
    exptime : `float`
        Exposure time for visits.
    nexp : `int`
        Number of exposures per visit.
    n_obs_template : `dict` or None
        The number of observations to take every season in each band.
        If None, sets to 3 each. Default None.
    pair_time : `float`
        The ideal time between pairs (minutes). Default 33.
    season : float
        The length of season (i.e., how long before templates expire) (days)
    season_start_hour : `float`
        Hour angle limits to use when gathering templates.
        Default -4 (hours)
    season_end_hour : `float`
       Hour angle limits to use when gathering templates.
       Default +2 (hours)
    max_alt : `float`
        The maximum altitude to use for the Surveys.
        Typically for BlobSurveys this is set lower than the max available,
        to about 76 degrees, to avoid long dome slews near azimuth.
        This is masked separately from the `safety_masks`.
    m5_weight : `float`
        The weight for the 5-sigma depth difference basis function.
    footprint_weight : `float`
        The weight on the survey footprint basis function.
    slewtime_weight : `float`
        The weight on the slewtime basis function.
    stayband_weight : `float`
        The weight on basis function that tries to stay avoid band changes.
    template_weight : `float`
        The weight to place on getting image templates every season.
    repeat_weight : `float`
        Weight that enhances (+ values) or decreases (- values) the likelihood
        of revisiting the same pointing within a two-hour time gap.
        Default of 0 here lets this survey repeat visits but total time
        in operation is limited anyway.
    scheduled_respect : `float`
        Ensure that blobs don't start within this many minutes of scheduled
        observations (from a ScriptedSurvey).
    night_pattern : `list` [ `bool` ]
        Which nights to let the survey execute (should be the opposite of
        the pattern for the NEO twilight survey).
        Default of [True, True] executes every night.
    science_program : `str`
        The science_program to use for visits from these surveys.
    blob_survey_params : `dict` or None
        A dictionary of additional kwargs to pass to the BlobSurvey.
        In particular, the times for typical slews, readtime, etc. are
        useful for setting the number of pointings to schedule within
        pair_time.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.
    pair_pad : `float`
        The padding to use for shadow mask. Default 5 (minutes).
    """

    if blob_survey_params is None:
        blob_survey_params = BLOB_SURVEY_PARAMS_DEFAULTS
    if safety_mask_params is None:
        safety_mask_params = {"nside": nside}
    else:
        safety_mask_params = copy.deepcopy(safety_mask_params)

    shadow_minutes = pair_time * 2 + pair_pad
    if shadow_minutes not in safety_mask_params or safety_mask_params["shadow_minutes"] < shadow_minutes:
        safety_mask_params["shadow_minutes"] = shadow_minutes

    surveys = []

    if n_obs_template is None:
        n_obs_template = {"u": 3, "g": 3, "r": 3, "i": 3, "z": 3, "y": 3}

    times_needed = [pair_time, pair_time * 2]
    for bandname, bandname2 in zip(band1s, band2s):
        detailer_list = []
        detailer_list.append(
            detailers.CameraRotDetailer(min_rot=np.min(camera_rot_limits), max_rot=np.max(camera_rot_limits))
        )
        detailer_list.append(detailers.CloseAltDetailer())
        detailer_list.append(detailers.FlushForSchedDetailer())
        detailer_list.append(detailers.LabelRegionsAndDDFs())
        # List to hold tuples of (basis_function_object, weight)
        bfs = []

        bfs.extend(
            standard_bf(
                nside,
                bandname=bandname,
                bandname2=bandname2,
                m5_weight=m5_weight,
                footprint_weight=footprint_weight,
                slewtime_weight=slewtime_weight,
                stayband_weight=stayband_weight,
                footprints=footprints,
                season=season,
                season_start_hour=season_start_hour,
                season_end_hour=season_end_hour,
            )
        )

        # XXX--magic numbers. Kinda nebulous why here.
        bfs.append(
            (
                bf.VisitRepeatBasisFunction(
                    gap_min=0, gap_max=2 * 60.0, bandname=None, nside=nside, npairs=20
                ),
                repeat_weight,
            )
        )

        # Make sure we respect scheduled observations
        bfs.append((bf.TimeToScheduledBasisFunction(time_needed=scheduled_respect), 0))

        bfs.append(
            (
                bf.AltAzShadowMaskBasisFunction(
                    nside=nside,
                    shadow_minutes=shadow_minutes,
                    max_alt=max_alt,
                    pad=3.0,
                ),
                0.0,
            )
        )
        if bandname2 is None:
            time_needed = times_needed[0]
        else:
            time_needed = times_needed[1]
        bfs.append((bf.TimeToTwilightBasisFunction(time_needed=time_needed, alt_limit=12), 0.0))

        # Let's turn off twilight blobs on nights where we are
        # doing NEO hunts
        bfs.append((bf.NightModuloBasisFunction(pattern=night_pattern), 0))

        # Add safety masks
        masks = safety_masks(**safety_mask_params)
        for m in masks:
            bfs.append((m, 0))

        # unpack the basis functions and weights
        weights = [val[1] for val in bfs]
        basis_functions = [val[0] for val in bfs]

        # Set survey_name and observation_reason
        if bandname2 is None:
            survey_name = "pair_%i, %s" % (pair_time, bandname)
        else:
            survey_name = "pair_%i, %s%s" % (pair_time, bandname, bandname2)
        if bandname2 is not None:
            detailer_list.append(detailers.TakeAsPairsDetailer(bandname=bandname2))

        observation_reason = f"pairs_{bandname}"
        if bandname2 is not None:
            observation_reason += f"{bandname2}"
        observation_reason += f"_{pair_time:.1f}"

        surveys.append(
            BlobSurvey(
                basis_functions,
                weights,
                bandname1=bandname,
                bandname2=bandname2,
                exptime=exptime,
                ideal_pair_time=pair_time,
                survey_name=survey_name,
                science_program=science_program,
                observation_reason=observation_reason,
                ignore_obs=ignore_obs,
                nexp=nexp,
                detailers=detailer_list,
                **blob_survey_params,
            )
        )

    return surveys


def generate_twilight_near_sun(
    nside: int = DEFAULT_NSIDE,
    night_pattern: list[bool] | None = None,
    nexp: int = 1,
    exptime: float = 15,
    ideal_pair_time: float = 5.0,
    max_airmass: float = 2.0,
    camera_rot_limits: tuple[float, float] = CAMERA_ROT_LIMITS,
    time_needed: float = 10.0,
    footprint_mask: npt.NDArray | float = 1,
    footprint_weight: float = 0.1,
    slewtime_weight: float = 3.0,
    stayband_weight: float = 3.0,
    band_dist_weight: float = 0.3,
    min_area: float | None = None,
    bands: str = "riz",
    n_repeat: int = 4,
    sun_alt_limit: float = -14.8,
    time_to_12deg: float = 25.0,
    slew_estimate: float = 4.5,
    max_elong: float = 60.0,
    ignore_obs: list[str] = ["DD", "pair", "long", "blob", "greedy", "template", "ToO"],
    science_program: str = SCIENCE_PROGRAM,
    safety_mask_params: dict | None = None,
) -> list[BlobSurvey]:
    """Generate a survey for observing NEO objects in twilight.

    This aims to cover ecliptic area within the LSST footprint, near the sun.

    Parameters
    ----------
    nside : `int`
        The HEALpix nside to use. Default to DEFAULT_NSIDE.
    night_pattern : `list` of `bool`
        A list of bools that set when the survey will be
        active. e.g., [True, False] for every-other night,
        [True, False, False] for every third night.
        Default None.
    nexp : `int`
        Number of snaps in a visit. Default 1.
    exptime : `float`
        Exposure time of visits. Default 15.
    ideal_pair_time : `float`
        Ideal time between repeat visits (minutes).
        Default 5
    max_airmass : `float`
        Maximum airmass to attempt (unitless). Default 2.
    camera_rot_limits : `list` of `float`
        The camera rotation limits to use (degrees).
        Default [-80, 80].
    time_needed : `float`
        How much time should be available
        (e.g., before twilight ends) (minutes).
        Default 10
    footprint_mask : `np.ndarray` or `float`
        Mask to apply to the constructed ecliptic target mask (None).
        Keeps ecliptic area within this region.
        Default None
    footprint_weight : `float`
        Weight for footprint basis function. Default 0.1 (unitless).
    slewtime_weight : `float`
        Weight for slewtime basis function. Default 3 (unitless)
    stayband_weight : `float`
        Weight for staying in the same band basis function.
        Default 3 (unitless)
    band_dist_weight : `float`
        Weight for the basis function that tries to keep the distribution
        of visits across filters even.
    min_area : `float`
        The area that needs to be available before the survey will return
        observations (sq degrees). Default None.
    bands : `str`
        The bands to use, default 'riz'
    n_repeat : `int`
        The number of times a blob should be repeated, default 4.
    sun_alt_limit : `float`
        Do not start unless sun is higher than this limit (degrees).
        Default -14.8.
    time_to_sunrise : `float`
        Do not execute if time to sunrise is greater than (minutes).
        Default 25.
    slew_estimate : `float`
        An estimate of how long it takes to slew between
        neighboring fields (seconds). Default 4.5
    max_elong : `float`
        Maximum solar elongation to mini-allow survey to reach.
    ignore_obs : `str` or `list` [ `str` ]
        Ignore observations by surveys that include the given substring(s).
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.
    """
    if safety_mask_params is None:
        safety_mask_params = {"nside": nside}
    else:
        safety_mask_params = copy.deepcopy(safety_mask_params)
    if (
        "shadow_minutes" not in safety_mask_params
        or safety_mask_params["shadow_minutes"] < ideal_pair_time * 4
    ):
        # pair_time * 4 (for the quad)
        safety_mask_params["shadow_minutes"] = ideal_pair_time * 4

    survey_name = "twilight_near_sun"
    observation_reason = "twilight_near_sun"
    footprint = ecliptic_area(nside=nside, mask=footprint_mask)
    constant_fp = ConstantFootprint(nside=nside)
    for bandname in bands:
        constant_fp.set_footprint(bandname, footprint)

    surveys = []
    for bandname in bands:
        detailer_list = []
        detailer_list.append(
            detailers.CameraRotDetailer(min_rot=np.min(camera_rot_limits), max_rot=np.max(camera_rot_limits))
        )
        detailer_list.append(detailers.CloseAltDetailer())
        # Should put in a detailer so things start at lowest altitude
        detailer_list.append(detailers.TwilightTripleDetailer(slew_estimate=slew_estimate, n_repeat=n_repeat))
        detailer_list.append(detailers.RandomBandDetailer(bands=bands))
        detailer_list.append(detailers.LabelRegionsAndDDFs())
        bfs = []

        bfs.append(
            (
                bf.FootprintBasisFunction(
                    bandname=bandname,
                    footprint=constant_fp,
                    out_of_bounds_val=np.nan,
                    nside=nside,
                ),
                footprint_weight,
            )
        )

        bfs.append(
            (
                bf.SlewtimeBasisFunction(bandname=bandname, nside=nside),
                slewtime_weight,
            )
        )
        bfs.append((bf.StrictBandBasisFunction(bandname=bandname), stayband_weight))
        bfs.append((bf.BandDistBasisFunction(bandname=bandname), band_dist_weight))
        # Need a toward the sun, reward high airmass, with an
        # airmass cutoff basis function.
        bfs.append(
            (
                bf.NearSunHighAirmassBasisFunction(nside=nside, max_airmass=max_airmass),
                0,
            )
        )

        bfs.append((bf.BandLoadedBasisFunction(bandnames=bandname), 0))
        bfs.append(
            (
                bf.SolarElongationMaskBasisFunction(min_elong=0.0, max_elong=max_elong, nside=nside),
                0,
            )
        )

        bfs.append((bf.NightModuloBasisFunction(pattern=night_pattern), 0))
        # Do not attempt unless the sun is getting high
        bfs.append(
            (
                bf.CloseToTwilightBasisFunction(
                    max_sun_alt_limit=sun_alt_limit, max_time_to_12deg=time_to_12deg
                ),
                0,
            )
        )

        # Add safety masks
        masks = safety_masks(**safety_mask_params)
        for m in masks:
            bfs.append((m, 0))

        # unpack the basis functions and weights
        weights = [val[1] for val in bfs]
        basis_functions = [val[0] for val in bfs]

        # Set huge ideal pair time and use the detailer to cut down
        # the list of observations to fit twilight?
        surveys.append(
            BlobSurvey(
                basis_functions,
                weights,
                bandname1=bandname,
                bandname2=None,
                ideal_pair_time=ideal_pair_time,
                nside=nside,
                exptime=exptime,
                survey_name=survey_name,
                ignore_obs=ignore_obs,
                dither="night",
                nexp=nexp,
                detailers=detailer_list,
                twilight_scale=False,
                area_required=min_area,
                observation_reason=observation_reason,
                science_program=science_program,
            )
        )
    return surveys



__all__ = ("gen_too_surveys",)

from copy import deepcopy

import numpy as np
import numpy.typing as npt

import rubin_scheduler.scheduler.detailers as detailers
from rubin_scheduler.scheduler.detailers import BandPickToODetailer
from rubin_scheduler.scheduler.surveys import ToOScriptedSurvey
from rubin_scheduler.utils import DEFAULT_NSIDE

from .lsst_surveys import EXPTIME, NEXP, SCIENCE_PROGRAM, safety_masks


def gen_too_surveys(
    nside: int = DEFAULT_NSIDE,
    detailer_list: list[detailers.BaseDetailer] | None = None,
    too_footprint: npt.NDArray | None = None,
    long_exp_nsnaps: int = 2,
    n_snaps: int = NEXP,
    science_program: str = SCIENCE_PROGRAM,
    safety_mask_params: dict | None = None,
) -> list[ToOScriptedSurvey]:
    """Generate a list of ToO surveys to follow up
    events passed in Conditions.

    Parameters
    ----------
    nside : `int`
        The HEALpix nside to use. Default to DEFAULT_NSIDE.
    detailer_list : `list` of `detailers.BaseDetailer`
        List of survey detailers.
    too_footprint : `np.ndarray` or None
        Footprint to contain ToOs within (such as the lsst footprint).
    long_exp_nsnaps : `int`
        The number of snaps for longer exposures. (60s??)
    n_snaps : `int`
        The number of snaps per visit for other exposures. (??)
    science_program : `str`
        Metadata to identify the science program for the visit.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to mass to the standard safety masks.

    Returns
    -------
    too_surveys : `list` [ `ToOScriptedSurvey` ]
        A list of ToO surveys configured to trigger a pre-specified sequence
        of visits in response to ToO events in the Conditions objects.
    """
    if safety_mask_params is None:
        safety_mask_params = {}
        safety_mask_params["nside"] = nside
    else:
        safety_mask_params = deepcopy(safety_mask_params)
    # No value of shadow_minutes with ToO surveys?
    masks = safety_masks(**safety_mask_params)

    too_surveys = []

    ############
    # Generic GW followup
    ############

    # XXX---There's some extra stuff about do different things
    # if there's limited time before it sets. Let's see if this can
    # work first

    # XXX--instructions say do 4th night only 1/3 of the time.
    # Just leaving off for now

    times = np.array([0, 24, 48], float)
    bands_at_times = ["ugrizy", "ugrizy", "ugrizy"]
    nvis = [3, 1, 1]
    exptimes = [120.0, 120.0, 120.0]
    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["GW_case_A"],
            survey_name="ToO, GW_case_A",
            flushtime=48.0,
            n_snaps=long_exp_nsnaps,
            # Update target_name to match the alert event ID
            target_name_base="GW_case_A",
            observation_reason="too_gw_case_a",
            science_program=science_program,
        )
    )

    ############
    # GW gold and GW unidentified gold
    ############

    times = np.array([0, 2, 4, 24, 48, 72], float)
    bands_at_times = ["gri", "gri", "gri", "ri", "ri", "ri"]
    nvis = [4, 4, 4, 6, 6, 6]
    exptimes = [30.0, 30.0, 30.0, 30.0, 30.0, 30.0]
    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["GW_case_B", "GW_case_C"],
            survey_name="ToO, GW_case_B_C",
            # Update target_name to match the alert event ID
            target_name_base="GW_case_B_C",
            observation_reason="too_gw_case_b_c",
            science_program=science_program,
            flushtime=48,
            n_snaps=long_exp_nsnaps,
        )
    )

    ############
    # GW silver and GW unidentified silver
    ############

    times = np.array([0, 24, 48, 72], float)
    bands_at_times = ["gi", "gi", "gi", "gi"]
    nvis = [1, 4, 4, 4]
    exptimes = [30.0, 30.0, 30.0, 30.0]
    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["GW_case_D", "GW_case_E"],
            survey_name="ToO, GW_case_D_E",
            # Update target_name to match the alert event ID
            target_name_base="GW_case_D_E",
            observation_reason="too_gw_case_d_e",
            science_program=science_program,
            flushtime=48,
            n_snaps=long_exp_nsnaps,
            event_gen_detailers=None,
        )
    )

    ############
    # BBH hole-black hole GW merger
    # If nearby (dist < 2200 Mpc) and dark time, use ugi
    # If distant (dist > 2200 Mpc) and dark time, use rgi
    # If bright time, use rzi
    ############

    # XXX--only considering bright objects now.

    # SM-- adding support for different BBH cases. We should have
    # a discussion about how to differentiate these, and if it is
    # possible to do so through the alert stream.

    event_detailers = [
        BandPickToODetailer(
            band_start="z",
            band_end="g",
            distance_limit=30e10,
            check_mounted=True,
            require_dark=True,
        ),
        BandPickToODetailer(
            band_start="r",
            band_end="u",
            distance_limit=2.2e6,
            check_mounted=True,
            require_dark=True,
        ),
    ]
    times = np.array([0, 2, 7, 9, 39], float) * 24
    bands_at_times = ["rzi"] * times.size
    nvis = [1] * times.size
    exptimes = [EXPTIME] * times.size

    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["BBH_case_A", "BBH_case_B", "BBH_case_C"],
            survey_name="ToO, BBH",
            # Update target_name to match the alert event ID
            target_name_base="BBH",
            observation_reason="too_bbh",
            science_program=science_program,
            flushtime=48,
            n_snaps=n_snaps,
            event_gen_detailers=event_detailers,
        )
    )

    ############
    # Lensed BNS
    ############

    times = np.array([1.0, 1.0, 25, 25, 49, 49], float)
    bands_at_times = ["g", "r"] * 3
    nvis = [1, 3] * 3
    exptimes = [EXPTIME, EXPTIME] * 3

    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["lensed_BNS_case_A"],
            survey_name="ToO, LensedBNS_A",
            # Update target_name to match the alert event ID
            target_name_base="LensedBNS_A",
            observation_reason="too_lensed_bns_a",
            science_program=science_program,
            flushtime=48.0,
            n_snaps=n_snaps,
        )
    )

    # This is the small skymap (15 deg^2 case)
    times = np.array([1.0, 1.0, 25, 25, 49, 49], float)
    bands_at_times = ["g", "r"] * 3
    nvis = [180, 120] * 3
    exptimes = [30] * times.size

    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["lensed_BNS_case_B"],
            survey_name="ToO, LensedBNS_B",
            # Update target_name to match the alert event ID
            target_name_base="LensedBNS_B",
            observation_reason="too_lensed_bns_b",
            science_program=science_program,
            flushtime=48.0,
            n_snaps=long_exp_nsnaps,
        )
    )

    ############
    # Neutrino detector followup
    ############

    times = np.array([0.0, 0.0, 15.0 / 60.0, 0.0, 24.0, 24.0, 144.0, 144.0], float)
    bands_at_times = ["u", "g", "r", "z", "g", "r", "g", "rz"]
    exptimes = [
        30,
        30,
        EXPTIME,
        EXPTIME,
        30,
        EXPTIME,
        30,
        EXPTIME,
    ]
    nvis = [1, 4, 1, 1, 4, 1, 4, 1]

    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["neutrino"],
            survey_name="ToO, neutrino",
            # Update target_name to match the alert event ID
            target_name_base="neutrino",
            observation_reason="too_neutrino",
            science_program=science_program,
            flushtime=20 * 24,
            n_snaps=n_snaps,
        )
    )

    ############
    # Solar System
    ############
    # For the solar system objects, probably want a custom survey object,
    # but this should work for now. Want to add a detailer to add a dither
    # position.

    times = np.array([0, 33 / 60.0, 66 / 60.0], float)
    bands_at_times = ["r"] * 3
    nvis = [1] * 3
    exptimes = [EXPTIME] * 3

    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["SSO_night"],
            survey_name="ToO, SSO_night",
            # Update target_name to match the alert event ID
            target_name_base="SSO_night",
            observation_reason="too_sso_general",
            science_program=science_program,
            flushtime=3.0,
            n_snaps=n_snaps,
        )
    )

    times = np.array([0, 10 / 60.0, 20 / 60.0])
    bands_at_times = ["z"] * 3
    nvis = [2] * 3
    exptimes = [15.0] * 3

    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["SSO_twilight"],
            survey_name="ToO, SSO_twi",
            # Update target_name to match the alert event ID
            target_name_base="SSO_twi",
            observation_reason="too_sso_twi",
            science_program=science_program,
            flushtime=3.0,
            n_snaps=n_snaps,
        )
    )

    ############
    # Galactic Supernova
    ############
    # For galactic supernovae, we want to tile continuously
    # in the region in 1s and 15s exposures for i band, until
    # a counterpart is identified

    times = np.array([0, 0, 0, 0] * 4, float)
    bands_at_times = ["i", "i", "i", "i"] * 4
    nvis = [1, 1, 1, 1] * 4
    exptimes = [1, 15, 1, 15] * 4

    too_surveys.append(
        ToOScriptedSurvey(
            masks,
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=deepcopy(detailer_list),
            too_types_to_follow=["SN_Galactic"],
            survey_name="ToO, galactic SN",
            # Update target_name to match the alert event ID
            target_name_base="SN_Galactic",
            observation_reason="too_sn_galactic",
            science_program=science_program,
            flushtime=48.0,
            n_snaps=n_snaps,
        )
    )

    return too_surveys



__all__ = ("define_ddf_seq", "gen_ddf_surveys")

import copy
import logging
import os
from pathlib import Path

import numpy as np
import pandas as pd

import rubin_scheduler.scheduler.detailers as detailers
import rubin_scheduler.scheduler.example.lsst_ddf_presched as ddf_presched
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler.example.lsst_surveys import (
    EXPTIME,
    NEXP,
    SCIENCE_PROGRAM,
    U_EXPTIME,
    U_NEXP,
    safety_masks,
)
from rubin_scheduler.scheduler.surveys import ScriptedSurvey
from rubin_scheduler.scheduler.utils import ScheduledObservationArray
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD

logger = logging.getLogger(__name__)


def define_ddf_seq() -> pd.DataFrame:
    """Define the sequences for each field"""

    short_squences = [
        {
            "u": 3,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 33,
        },
        {
            "y": 2,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 33,
        },
        {
            "g": 2,
            "i": 2,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 56,
            "even_odd": "even",
        },
        {
            "r": 2,
            "z": 2,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 56,
            "even_odd": "odd",
        },
    ]

    shallow_squences = [
        {
            "u": 3,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 75,
        },
        {
            "y": 2,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 75,
        },
        {
            "g": 2,
            "i": 2,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 100,
            "even_odd": "even",
        },
        {
            "r": 2,
            "z": 2,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 100,
            "even_odd": "odd",
        },
    ]

    deep_sequences = [
        {
            "u": 8,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 75,
        },
        {
            "y": 20,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 75,
        },
        {
            "g": 2,
            "i": 2,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 200,
            "even_odd": "even",
        },
        {
            "r": 2,
            "z": 2,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 200,
            "even_odd": "odd",
        },
        {
            "g": 4,
            "r": 18,
            "i": 55,
            "z": 52,
            "season_length": 180,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 110,
        },
    ]

    euclid_deep_seq = [
        {
            "u": 30,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 75,
        },
        {
            "y": 40,
            "season_length": 225,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 75,
        },
        {
            "g": 4,
            "i": 4,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 200,
            "even_odd": "even",
        },
        {
            "r": 4,
            "z": 4,
            "season_length": 225,
            "flush_length": 0.5,
            "g_depth_limit": 22.8,
            "n_sequences": 200,
            "even_odd": "odd",
        },
        {
            "g": 8,
            "r": 36,
            "i": 110,
            "z": 104,
            "season_length": 125,
            "flush_length": 2.0,
            "g_depth_limit": 23.5,
            "n_sequences": 75,
        },
    ]

    short_seasons = {
        "XMM_LSS": [0, 10],
        "ELAISS1": [0, 10],
        "ECDFS": [0, 10],
        "EDFS_a": [0, 10],
    }

    shallow_seasons = {
        "COSMOS": [0, 4, 5, 6, 7, 8, 9, 10],
        "XMM_LSS": [1, 2, 3, 5, 6, 7, 8, 9],
        "ELAISS1": [1, 2, 3, 4, 6, 7, 8, 9],
        "ECDFS": [1, 2, 3, 4, 5, 7, 8, 9],
        "EDFS_a": [2, 3, 4, 5, 6, 7, 8, 9],
    }

    deep_seasons = {
        "COSMOS": [1, 2, 3],
        "XMM_LSS": [4],
        "ELAISS1": [5],
        "ECDFS": [6],
        "EDFS_a": [1],
    }

    dataframes = []

    for ddf_name in short_seasons:
        for season in short_seasons[ddf_name]:
            dict_for_df = {
                "ddf_name": ddf_name,
                "season": season,
                "even_odd": "None",
            }
            for key in "ugrizy":
                dict_for_df[key] = 0

            for seq in short_squences:
                row = copy.copy(dict_for_df)
                for key in seq:
                    row[key] = seq[key]
                dataframes.append(pd.DataFrame.from_dict(row, orient="index").T)

    for ddf_name in shallow_seasons:
        for season in shallow_seasons[ddf_name]:
            dict_for_df = {
                "ddf_name": ddf_name,
                "season": season,
                "even_odd": "None",
            }
            for key in "ugrizy":
                dict_for_df[key] = 0

            for seq in shallow_squences:
                row = copy.copy(dict_for_df)
                for key in seq:
                    row[key] = seq[key]
                dataframes.append(pd.DataFrame.from_dict(row, orient="index").T)

    for ddf_name in deep_seasons:
        for season in deep_seasons[ddf_name]:
            dict_for_df = {
                "ddf_name": ddf_name,
                "season": season,
                "even_odd": "None",
            }
            for key in "ugrizy":
                dict_for_df[key] = 0
            if ddf_name == "EDFS_a":
                for seq in euclid_deep_seq:
                    row = copy.copy(dict_for_df)
                    for key in seq:
                        row[key] = seq[key]
                    dataframes.append(pd.DataFrame.from_dict(row, orient="index").T)
            else:
                for seq in deep_sequences:
                    row = copy.copy(dict_for_df)
                    for key in seq:
                        row[key] = seq[key]

                    dataframes.append(pd.DataFrame.from_dict(row, orient="index").T)

    result = pd.concat(dataframes)

    return result


def gen_ddf_surveys(
    detailer_list: list[detailers.BaseDetailer] | None = None,
    nside: int = DEFAULT_NSIDE,
    expt: dict | None = None,
    nexp: dict | None = None,
    survey_start: float = SURVEY_START_MJD,
    survey_length: int = 10,
    survey_name: str = "deep drilling",
    science_program: str = SCIENCE_PROGRAM,
    shadow_minutes: float = 30,
    save: bool = True,
    save_filename: str = "example_ddf_array.npz",
    save_path: str = None,
    safety_mask_params: dict | None = None,
) -> list[ScriptedSurvey]:
    """Generate surveys for DDF observations.

    Parameters
    ----------
    detailer_list : `list` [ `rubin_scheduler.scheduler.Detailer` ]
        Detailers for DDFs. Default None.
    nside : `int`
        Nside for the survey. Used for mask basis functions.
    expt : `dict`  { `str` : `float` } or None
        Exposure time for DDF visits.
        Default of None uses defaults of EXPTIME/U_EXPTIME.
    nexp : `dict` { `str` : `int` } or None
        Number of exposures per visit.
        Default of None uses defaults of NEXP/U_NEXP.
    survey_start : `float`
        Start MJD of the survey. Used for prescheduling DDF visits.
    survey_length : `float`
        Length of the survey. Used for prescheduling DDF visits.
        In years.
    science_program : `str`
        Name of the science program for the Survey.
    shadow_minutes : `float`
        The expected length of time for a series of DDF visits
        to execute. Masks area on the sky which will enter unobservable
        regions within shadow_minutes.
    save : `bool`
        Save the resulting ddf array for faster restore next time run.
    save_filename : `str`
        Filename of the saved ddf array.
    save_path : `str`
        Path to saved DDF file. If none, uses get_data_dir to look for it.
    safety_mask_params : `dict` or None
        A dictionary of additional kwargs to ass to the standard safety masks.

    Returns
    -------
    ddf_surveys : `list` [ `ScriptedSurvey` ]
        A list of Scripted surveys configured with pre-scheduled DDF visits.
    """
    if safety_mask_params is None:
        safety_mask_params = {}
        safety_mask_params["nside"] = nside
    else:
        safety_mask_params = copy.deepcopy(safety_mask_params)
    if "shadow_minutes" not in safety_mask_params or safety_mask_params["shadow_minutes"] < shadow_minutes:
        safety_mask_params["shadow_minutes"] = shadow_minutes

    if expt is None:
        expt = {
            "u": U_EXPTIME,
            "g": EXPTIME,
            "r": EXPTIME,
            "i": EXPTIME,
            "z": EXPTIME,
            "y": EXPTIME,
        }
    if nexp is None:
        nexp = {"u": U_NEXP, "g": NEXP, "r": NEXP, "i": NEXP, "z": NEXP, "y": NEXP}

    # Potential pre-computed obs_array:
    if save_path is None:
        save_path = Path(get_data_dir(), "scheduler")
    # Potetial pre-computed obs_array:
    pre_comp_file = Path(save_path, save_filename)

    # Hash of the files that define the DDF sequences, to identify
    # if the saved file comes from the same sources.
    hash_digest = ddf_presched.calculate_checksum([__file__, ddf_presched.__file__])
    passed_kwargs = {
        "expt": expt,
        "nexp": nexp,
        "survey_start": survey_start,
        "survey_length": survey_length,
        "science_program": science_program,
    }

    # Always try to load the pre-computed data.
    obs_array = None
    if os.path.exists(pre_comp_file):
        loaded = np.load(pre_comp_file, allow_pickle=True)
        if loaded["hash_digest"] == hash_digest:
            # Check that all the kwargs match
            kwargs_match = True
            for key in passed_kwargs:
                if passed_kwargs[key] != loaded[key]:
                    kwargs_match = False

            if kwargs_match:
                logger.info("Loading DDF array from %s" % pre_comp_file)
                obs_array_loaded = loaded["obs_array"]
                # Convert back to a full ScheduledObservationArray?
                obs_array = ScheduledObservationArray(obs_array_loaded.size)
                for key in obs_array_loaded.dtype.names:
                    obs_array[key] = obs_array_loaded[key]
        loaded.close()

    if obs_array is None:
        logger.info("Generating DDF array")
        ddf_dataframe = define_ddf_seq()

        obs_array = ddf_presched.generate_ddf_scheduled_obs(
            ddf_dataframe,
            expt=expt,
            nsnaps=nexp,
            survey_start_mjd=survey_start,
            survey_length=survey_length,
            science_program=science_program,
        )
        # Save computation for later
        if save:
            logger.info("Saving DDF array to %s" % pre_comp_file)
            np.savez(
                pre_comp_file,
                obs_array=obs_array.view(np.ndarray),
                hash_digest=hash_digest,
                expt=expt,
                nexp=nexp,
                survey_start=survey_start,
                survey_length=survey_length,
                science_program=science_program,
            )

    survey1 = ScriptedSurvey(
        safety_masks(**safety_mask_params),
        nside=nside,
        detailers=detailer_list,
        survey_name=survey_name,
        before_twi_check=False,
    )
    survey1.set_script(obs_array)

    return [survey1]



__all__ = ("SummitWrapper",)

import copy
import warnings

import numpy as np

from rubin_scheduler.scheduler.utils import IntRounded
from rubin_scheduler.utils import _ra_dec2_hpid


class SummitWrapper:
    """Wrap a core scheduler so observations can be requested
    and are assumed to be completed.

    Parameters
    ----------
    core_scheduler : `rubin_scheduler.scheduler.schedulers.CoreScheduler`
        A CoreScheduler object.
    """

    def __init__(self, core_scheduler):

        # A standard CoreScheduler object that
        # is only given observations that have been
        # completed.
        self.core_scheduler = core_scheduler
        self.conditions = None
        self.clear_ahead()

        # For compatibility with sim_runner.
        # Not tracking flushed observations
        self.flushed = 0

    def clear_ahead(self):
        """Reset the ahead scheduler and pending observations"""

        # The ahead_scheduler acts as a copy of
        # core scheduler, but will assume any requested
        # observations are completed instantly, so requesting
        # multiple observations does not risk getting the same
        # observation multiple times.
        self.ahead_scheduler = copy.deepcopy(self.core_scheduler)
        self.requested_but_unadded_ids = []
        self.requested_but_unadded_obs = []
        # Have we added observations so
        # self.core_scheduler and self.ahead_scheduler
        # are out of sync?
        self.need_replay = False

    def flush_queue(self):
        self.core_scheduler.flush_queue()
        self.clear_ahead()

    def add_observation(self, observation):
        """Add actually completed observations."""

        self.core_scheduler.add_observation(observation)

        # Assume everything up to the ID has been observed.
        # Should be ok to add out of order, as long as everything up
        # to the largest ID is added before request_observation is called.
        indx = np.searchsorted(
            self.requested_but_unadded_ids, observation["target_id"].view(np.ndarray).max(), side="right"
        )

        # Should this delete everything up to the indx?
        # I think that's ok if we assume things will get added in order
        if indx.size > 0:
            self.requested_but_unadded_ids = self.requested_but_unadded_ids[indx:]
            self.requested_but_unadded_obs = self.requested_but_unadded_obs[indx:]

        self.need_replay = True

    def _check_queue_mjd_only(self, mjd):
        """
        Check if there are things in the queue that can be executed
        using only MJD and not full conditions.
        This is primarily used by sim_runner to reduce calls calculating
        updated conditions when they are not needed.
        """
        result = False
        if np.sum(self.ahead_scheduler.queue_manager.need_observing) > 0:
            fb_mjd = self.ahead_scheduler.return_active_queue()["flush_by_mjd"]
            not_flushed_yet = np.where(IntRounded(mjd) < IntRounded(fb_mjd))[0]
            no_flush_set = np.where(fb_mjd == 0)[0]
            if (np.size(not_flushed_yet) > 0) | (np.size(no_flush_set) > 0):
                result = True
        return result

    def update_conditions(self, conditions):

        self.conditions = conditions
        self.ahead_scheduler.update_conditions(conditions)
        self.core_scheduler.update_conditions(conditions)

    def _fill_obs_values(self, observation):
        """Fill in values of an observation assuming it will be
        observed approximately now.
        """

        # Nearest neighbor from conditions maps
        hpid = _ra_dec2_hpid(self.conditions.nside, observation["RA"], observation["dec"])

        # XXX--Need to go through and formally list the columns
        # that are minimally required for adding observations to Features.
        # This seems to work, but nothing stopping someone from asking for
        # something random like moon phase and then it will fail. Maybe
        # should set all un-defined columns to np.nan so it's clear they
        # can't be used unless this method is updated.
        observation["mjd"] = self.conditions.mjd
        observation["FWHMeff"] = self.conditions.fwhm_eff[observation["band"][0]][hpid]
        observation["airmass"] = self.conditions.airmass[hpid]
        observation["fivesigmadepth"] = self.conditions.m5_depth[observation["band"][0]][hpid]
        observation["night"] = self.conditions.night

        return observation

    def request_observation(self, mjd=None):
        """Request an observation, assuming previously requested
        observations were successfully observed.
        """

        if mjd is None:
            mjd = self.conditions.mjd

        # If observations have been added, need to
        # rebuild ahead_scheduler
        if self.need_replay:
            self.ahead_scheduler = copy.deepcopy(self.core_scheduler)
            for obs in self.requested_but_unadded_obs:
                if mjd < obs["mjd"]:
                    msg = f"""Adding observation with MJD={obs["mjd"]}, but think it is
                          currently MJD={mjd}, so we are adding observations
                          from the future. That seems like it shouldn't be possible.
                          """
                    warnings.warn(msg)
                self.ahead_scheduler.add_observation(obs)
            self.need_replay = False
            self.ahead_scheduler.update_conditions(self.conditions)

        # If we fill the queue, need to add that to the core scheduler
        # so it knows about it
        if self.ahead_scheduler.return_active_queue().size > 0:
            result_plain = self.ahead_scheduler.request_observation(mjd=mjd)
        else:
            # Now, we either refill the queue or
            # generate a one-off new observation and have an empty queue.
            result_plain = self.ahead_scheduler.request_observation(mjd=mjd)
            self.core_scheduler.append_to_queue(result_plain.copy())
            ahead_active_queue = self.ahead_scheduler.return_active_queue()
            if ahead_active_queue.size > 0:
                self.core_scheduler.append_to_queue(ahead_active_queue)

        obs_filled = self._fill_obs_values(result_plain.copy())
        self.requested_but_unadded_ids.append(obs_filled["target_id"].view(np.ndarray).max())
        self.requested_but_unadded_obs.append(obs_filled)

        # Add requested observation to the ahead
        # scheduler, so if we call request_observation again,
        # it will think this has been completed.
        self.ahead_scheduler.add_observation(obs_filled)

        return result_plain



import unittest

import healpy as hp
import numpy as np

import rubin_scheduler.scheduler.basis_functions as bf
from rubin_scheduler.scheduler.detailers import RotspUpdateDetailer
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.schedulers import BaseQueueManager
from rubin_scheduler.scheduler.utils import Footprint, ObservationArray
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD, hpid2_ra_dec


class TestQM(unittest.TestCase):

    def test_qm(self):
        observations = ObservationArray(n=3)
        observations["RA"] = np.radians(np.arange(0, 3, 90))
        observations["dec"] = np.radians([-20, -30, -40])
        observations["band"] = "r"
        observations["target_id"] = np.arange(3)

        conditions = Conditions()
        conditions.mjd = SURVEY_START_MJD + 10

        detailer = RotspUpdateDetailer()

        # make a footprint, mask it off below dec -30
        npix = hp.nside2npix(DEFAULT_NSIDE)
        fp = np.ones(hp.nside2npix(DEFAULT_NSIDE))
        ra, dec = hpid2_ra_dec(DEFAULT_NSIDE, np.arange(npix))
        fp[np.where(dec < -35)] = 0
        fp_obj = Footprint(SURVEY_START_MJD)
        fp_obj.set_footprint("r", fp)

        fp_bf = bf.FootprintBasisFunction(
            bandname="r",
            footprint=fp_obj,
            out_of_bounds_val=np.nan,
            nside=DEFAULT_NSIDE,
        )

        qm = BaseQueueManager(detailers=[detailer], basis_functions=[fp_bf])

        qm.set_queue(observations)

        # Do we get the active queue?
        active_queue = qm.return_active_queue()
        assert active_queue.size == 3

        one_obs = qm.request_observation(conditions)

        assert np.size(one_obs) == 1

        # Active queue should be the same
        active_queue = qm.return_active_queue()
        assert active_queue.size == 3

        qm.add_observation(one_obs)

        # Now should be down to 2
        active_queue = qm.return_active_queue()
        assert active_queue.size == 2

        qm.set_queue(observations)

        # Active queue should be reset
        active_queue = qm.return_active_queue()
        assert active_queue.size == 3

        # Does check for visibility
        all_valid_obs = qm.request_observation(conditions, whole_queue=True)
        assert np.size(all_valid_obs) < 3

        # Does check that should toss one for being out of footprint
        all_valid_obs = qm.request_observation(conditions, n_return=4)
        assert np.size(all_valid_obs) == 2

        # Check we can wipe out lots at once
        names = list(observations.dtype.names)
        types = [observations[name].dtype for name in names]
        names.append("hpid")
        types.append(int)
        ndt = list(zip(names, types))
        observations_hpid_in = np.zeros(3, dtype=ndt)
        qm.add_observations_array(observations, observations_hpid_in)
        # Should have matched everything
        active_queue = qm.return_active_queue()
        assert active_queue.size == 0

        qm.set_queue(observations)
        qm.flush_queue()
        active_queue = qm.return_active_queue()
        assert active_queue.size == 0

        # Test the detailer is doing something.
        # shouldn't be able to observe rotSkyPos
        # of 0 and 180, so at least one of these
        # should change.
        qm.set_queue(observations)
        result1 = qm.request_observation(conditions, n_return=2)

        rot_same_1 = np.all(result1["rotSkyPos"] == 0)

        observations["rotSkyPos"] = np.pi
        qm.set_queue(observations)
        result2 = qm.request_observation(conditions, n_return=2)

        rot_same_2 = np.all(result2["rotSkyPos"] == np.pi)

        assert rot_same_1 * rot_same_2 == 0


if __name__ == "__main__":
    unittest.main()



import unittest

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.example.too_surveys import gen_too_surveys
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.surveys import ToOScriptedSurvey
from rubin_scheduler.scheduler.utils import TargetoO
from rubin_scheduler.utils import hpid2_ra_dec


class TestToO(unittest.TestCase):

    def test_tesselation(self):
        """Test tesselation"""
        survey = ToOScriptedSurvey([])
        ra1, dec1 = survey._tesselate([100, 101])
        # This should spin things
        ra2, dec2 = survey._tesselate([100, 101])

        assert ra2[0] != ra1[0]

        # If we ask for a single HEALpix
        # Should only give a pointing at
        # one location

        ra1, dec1 = survey._tesselate([100])

        assert len(ra1) == 1

        # Should be no change if we call again
        ra2, dec2 = survey._tesselate([100])

        assert ra1 == ra2

    def test_positions(self):
        """Test tesselation rotates as expected"""

        nside = 32
        times = [1, 24, 48]
        bands_at_times = ["i", "i", "i"]
        nvis = [1, 1, 1]
        exptimes = [30, 30, 30]

        too_footprint = np.ones(hp.nside2npix(nside))

        survey = ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=[],
            too_types_to_follow=["test_too"],
            survey_name="ToO, test",
            flushtime=48.0,
            n_snaps=1,
            target_name_base="ToO_test",
            observation_reason="too_test",
        )

        # Generate a ToO event
        target_map = np.zeros(hp.nside2npix(nside))
        ra, dec = hpid2_ra_dec(nside, np.arange(target_map.size))
        target_map[np.where(dec < -85)[0]] = 1
        too_event = TargetoO(
            100, target_map, 6500, 100, ra_rad_center=None, dec_rad_center=None, too_type="test_too"
        )
        conditions = Conditions()
        conditions.targets_of_opportunity = [too_event]

        # Give ToO to survey
        survey.update_conditions(conditions)

        # All pointings should be unique
        n_unique = np.size(np.unique(survey.obs_wanted["RA"]))
        n_all = np.size(survey.obs_wanted)

        assert n_unique > 1
        assert n_unique == n_all

        # Check if we have multiple filters, they are at same position
        bands_at_times = ["ig", "ig", "ig"]
        survey = ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=[],
            too_types_to_follow=["test_too"],
            survey_name="ToO, test",
            flushtime=48.0,
            n_snaps=1,
            target_name_base="ToO_test",
            observation_reason="too_test",
        )

        survey.update_conditions(conditions)

        n_unique = np.size(np.unique(survey.obs_wanted["RA"]))
        n_all = np.size(survey.obs_wanted)

        assert n_unique > 1
        assert n_unique == n_all / 2

        # Check if we have multiple filters, and multiple nvis
        bands_at_times = ["ig", "ig", "ig"]
        nvis = [2, 2, 2]
        survey = ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=[],
            too_types_to_follow=["test_too"],
            survey_name="ToO, test",
            flushtime=48.0,
            n_snaps=1,
            target_name_base="ToO_test",
            observation_reason="too_test",
        )

        survey.update_conditions(conditions)

        n_unique = np.size(np.unique(survey.obs_wanted["RA"]))
        n_all = np.size(survey.obs_wanted)

        assert n_unique > 1
        assert n_unique == n_all / 2

        # Turn dither off
        survey = ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=[],
            too_types_to_follow=["test_too"],
            survey_name="ToO, test",
            flushtime=48.0,
            n_snaps=1,
            target_name_base="ToO_test",
            observation_reason="too_test",
            dither_per_visit=False,
        )

        # Give ToO to survey
        survey.update_conditions(conditions)

        # Now pointings should repeat
        n_unique = np.size(np.unique(survey.obs_wanted["RA"]))
        n_all = np.size(survey.obs_wanted)

        assert n_unique > 1
        assert n_unique < n_all

    def test_events(self):
        """Test the ToO events, make sure each one generates"""

        nside = 32
        too_footprint = np.ones(hp.nside2npix(nside))
        surveys = gen_too_surveys(nside=nside, too_footprint=too_footprint)

        conditions = Conditions()

        target_map = np.zeros(hp.nside2npix(nside))

        ra, dec = hpid2_ra_dec(nside, np.arange(target_map.size))

        target_map[np.where(dec < -85)[0]] = 1

        target_map_single = target_map * 0
        # for lensed_BNS_case_B, it is doing a bad loop, so need
        # to keep footprint very small.
        target_map_single[np.where(target_map == 1)[0].min()] = 1

        for survey in surveys:
            for type in survey.too_types_to_follow:
                tm = target_map
                if type == "lensed_BNS_case_B":
                    tm = target_map_single
                too = TargetoO(100, tm, 6500, 100, ra_rad_center=None, dec_rad_center=None, too_type=type)
                conditions.targets_of_opportunity = [too]

                survey.update_conditions(conditions)
                assert np.size(survey.obs_wanted.shape) > 0
                survey.obs_wanted = np.array([])

    def test_sort(self):
        """Test the sorting works ok"""

        nside = 32
        times = [1, 24, 48]
        bands_at_times = ["i", "i", "i"]
        nvis = [1, 1, 1]
        exptimes = [30, 30, 30]

        too_footprint = np.ones(hp.nside2npix(nside))

        # Check if we have multiple filters, and multiple nvis
        bands_at_times = ["ig", "ig", "ig"]
        nvis = [2, 2, 2]
        survey = ToOScriptedSurvey(
            [],
            nside=nside,
            followup_footprint=too_footprint,
            times=times,
            bands_at_times=bands_at_times,
            nvis=nvis,
            exptimes=exptimes,
            detailers=[],
            too_types_to_follow=["test_too"],
            survey_name="ToO, test",
            flushtime=48.0,
            n_snaps=1,
            target_name_base="ToO_test",
            observation_reason="too_test",
            return_n_limit=2,
        )

        # Generate a ToO event
        target_map = np.zeros(hp.nside2npix(nside))
        ra, dec = hpid2_ra_dec(nside, np.arange(target_map.size))
        target_map[np.where(dec < -85)[0]] = 1
        too_event = TargetoO(
            100, target_map, 6500, 100, ra_rad_center=None, dec_rad_center=None, too_type="test_too"
        )
        conditions = Conditions()
        conditions.targets_of_opportunity = [too_event]

        # Give ToO to survey
        survey.update_conditions(conditions)
        conditions.mjd = survey.mjd_start + 0.01
        # Sun and moon out of the way
        conditions.moon_ra = 0.0
        conditions.moon_dec = np.pi / 2
        conditions.sun_alt = -np.pi / 2.0

        conditions.tel_alt_limits = [-np.pi / 2, np.pi / 2]
        conditions.tel_az_limits = [-3 * np.pi, 3 * np.pi]
        conditions.mounted_bands = ["u", "g", "r", "i", "z", "y"]

        survey.obs_wanted["HA_min"] = 12
        survey.obs_wanted["HA_max"] = 12

        obs = survey._check_list(conditions)

        assert obs.size == 2

        assert np.size(np.unique(obs["band"])) == 1


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import SummitWrapper


class TestSummitWrapper(unittest.TestCase):

    def test_baseline(self):

        # Run the usual baseline
        mjd_start = utils.SURVEY_START_MJD
        nside = 32
        survey_length = 2.0  # days
        n_visit_limit = 200

        # Reset and run with a wrapped CoreScheduler
        scheduler_wrapped = SummitWrapper(example_scheduler(nside=nside, survey_start_mjd=mjd_start))
        observatory = ModelObservatory(
            nside=nside, mjd_start=mjd_start, cloud_data="ideal", downtimes="ideal"
        )
        observatory, scheduler_wrapped, observations_wrapped = sim_runner(
            observatory,
            scheduler_wrapped,
            sim_duration=survey_length,
            filename=None,
            n_visit_limit=n_visit_limit,
        )

        scheduler = example_scheduler(nside=nside, survey_start_mjd=mjd_start)
        observatory = ModelObservatory(
            nside=nside, mjd_start=mjd_start, cloud_data="ideal", downtimes="ideal"
        )
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None, n_visit_limit=n_visit_limit
        )

        # Should be the same number of unique target IDs
        assert np.size(np.unique(observations["target_id"])) == np.size(
            np.unique(observations_wrapped["target_id"])
        )

        # I don't mind if target_id got ahead on one method
        observations_wrapped["target_id"] = observations["target_id"]

        # Should be the same
        np.testing.assert_array_equal(observations, observations_wrapped)


if __name__ == "__main__":
    unittest.main()



__all__ = ("BaseQueueManager",)

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.utils import IntRounded, ObservationArray


class BaseQueueManager:
    """Class for managing a queue of desired observations.

    NOTE: If using masking basis functions, one should
    ensure Survey objects are using the same masks as the
    QueueManager, else one will end up in a loop where
    surveys propose visits that the QueueManager always rejects.
    E.g., if the surveys don't mask clouds, all their observations
    can be rejected by the QueueManager.

    Parameters
    ----------
    detailers : `list`
        List of detailers to apply to observations as they are
        selected from the queue. Mostly for setting camera
        rotation angles dynamically, or skipping observations
        that have become clouded out or gotten close to zenith.
    basis_functions : `list`
        List of basis functions. Assumed to be only used for
        masking, thus no weights for the basis functions.

    """

    def __init__(self, detailers=None, basis_functions=None):
        if detailers is None:
            self.detailers = []
        else:
            self.detailers = detailers
        if basis_functions is None:
            self.basis_functions = []
        else:
            self.basis_functions = basis_functions
        self.desired_observations_array = ObservationArray(n=0)
        # Array to track which desired_observations_array
        # have been observed
        self.need_observing = np.zeros(0, dtype=bool)

    def flush_queue(self):
        self.desired_observations_array = ObservationArray(n=0)
        self.need_observing = np.zeros(0, dtype=bool)

    def append_to_queue(self, observation_array):
        need_observing = np.ones(observation_array.size, dtype=bool)
        self.need_observing = np.concatenate([self.need_observing, need_observing])
        self.desired_observations_array = np.concatenate([self.desired_observations_array, observation_array])

    def set_queue(self, observation_array):
        self.desired_observations_array = observation_array
        self.need_observing = np.ones(observation_array.size, dtype=bool)

    def add_observation(self, observation, **kwargs):
        if self.desired_observations_array.size > 0:
            self.mark_observation_completed(observation)

        for bf in self.basis_functions:
            bf.add_observation(observation, **kwargs)
        for detailer in self.detailers:
            detailer.add_observation(observation, **kwargs)

    def add_observations_array(self, observation_array, observations_hpid_in):
        if self.desired_observations_array.size > 0:
            indx = np.isin(self.desired_observations_array["target_id"], observation_array["target_id"])
            if np.size(indx) > 0:
                self.need_observing[indx] = False

            good = np.isin(observations_hpid_in["ID"], observation_array["ID"])
            observations_hpid = observations_hpid_in[good]
            for bf in self.basis_functions:
                bf.add_observations_array(observation_array, observations_hpid)
            for detailer in self.detailers:
                detailer.add_observations_array(observation_array, observations_hpid)

    def compute_reward(self, conditions):
        reward = 0
        for bf in self.basis_functions:
            reward += bf(conditions)
        return reward

    def _check_queue_mjd_only(self, mjd, conditions):
        """
        Check if there are things in the queue that can be executed
        using only MJD and not full conditions.
        This is primarily used by sim_runner to reduce calls calculating
        updated conditions when they are not needed.
        """

        # With queue manager doing cloud dodging, need full conditions
        # now. Go ahead and use stale conditions, but update the
        # mjd so things get flushed if needed.
        valid_obs = self.find_valid_observations_indx(conditions, mjd=mjd)

        return np.size(valid_obs) > 0

    def find_valid_observations_indx(self, conditions, mjd=None):
        """Return the indices of desired_observations_array
        that can be observed
        """

        if mjd is None:
            mjd = conditions.mjd

        # Still valid date
        mjd_ok = np.where(
            (IntRounded(mjd) <= IntRounded(self.desired_observations_array["flush_by_mjd"]))
            | (self.desired_observations_array["flush_by_mjd"] == 0)
        )[0]

        valid = np.where(self.need_observing)[0]

        valid = np.intersect1d(mjd_ok, valid)

        # Compute reward from basis functions
        reward = self.compute_reward(conditions)
        # If reward is an array, then it's a HEALpy map and we
        # need to interpolate to the actual positions we want.
        # now to interpolate to the reward positions
        if np.size(reward) > 1:
            reward_interp = hp.get_interp_val(
                reward,
                np.degrees(self.desired_observations_array["RA"][valid]),
                np.degrees(self.desired_observations_array["dec"][valid]),
                lonlat=True,
            )
            sub_valid_reward = np.isfinite(reward_interp)
            valid = np.ravel(valid[sub_valid_reward])

        return valid

    def request_observation(self, conditions, whole_queue=False, n_return=1):
        """
        Parameters
        ----------
        conditions : ``
            Conditions object
        whole_queue : `bool`
            If True, return all desired observations that
            pass basis function checks. Default False.
        n_return : `int`
            Number of observations to return. Ignored
            if whole_queue is True.
        """

        indx = self.find_valid_observations_indx(conditions)
        if not whole_queue:
            if indx.size > n_return:
                indx = indx[0:n_return]

        result = self.desired_observations_array[indx].copy()
        if np.size(result) > 0:
            for det in self.detailers:
                result = det(result, conditions)

        return result

    def return_active_queue(self):
        """Return array of observations that are waiting to be executed"""
        return self.desired_observations_array[self.need_observing]

    def mark_observation_completed(self, observation):
        """Mark something in the queue as completed.
        Matches on target_id only.
        """
        match = np.where(self.desired_observations_array["target_id"] == observation["target_id"])[0]
        self.need_observing[match] = False
